#!/usr/bin/env python3

"""
Functions used to evaluate statistical significance of changes.
cmdoret, 20200403
"""
import numpy as np
import scipy.stats as ss


def vals_to_percentiles(vals: np.ndarray, dist: np.ndarray) -> np.ndarray:
    """
    Return the percentiles corresponding to values in a distribution.

    Parameters
    ----------
    vals : np.ndarray of floats
        The values for which percentiles should be returned.
    dist : np.ndarray of floats
        The distribution to use when computing the percentiles.

    Returns
    -------
    np.ndarray of floats :
        The array of percentiles corresponding to input values
    """
    # Get indices corresponding to vals in the distribution
    sorted_dist = np.sort(dist)
    percentiles = np.searchsorted(sorted_dist, vals) / len(sorted_dist)
    return percentiles


def vec_ttest(arr1, arr2):
    """
    Given 2 numpy arrays containing data for 2 different conditions, where
    each array has shape shape rxd where r is the number of replicates and N
    is the number of independent tests to perform. N must be equal in both array,
    however the number of replicates can differ.
    """
    if not (len(arr1.shape) == len(arr2.shape) == 2):
        raise ValueError("Both arrays must be 2-dimensional.")
    if arr1.shape[0] != arr2.shape[0]:
        raise ValueError(
            "Both arrays must have the same number of test datasets"
        )
    # Number of sample in each test (both conditions)
    n_samples = arr1.shape[1] + arr2.shape[1]
    # Compute mean and variance for each test independently
    means_1, means_2 = arr1.mean(axis=1), arr2.mean(axis=1)
    var1, var2 = arr1.var(axis=1, ddof=1), arr2.var(axis=1, ddof=1)
    stds = np.sqrt((var1 + var2) / 2)
    stds[stds == 0] = np.nan
    # alt - control
    t_stats = (means_2 - means_1) / (stds * np.sqrt(1 / n_samples))
    t_stats[np.isnan(t_stats)] = 0

    return t_stats


# TODO: Write a vectorized linear model function for cases with more than
# 2 (quantitative) conditions
def vec_lm():
    ...


def tvals_to_pvals(tvals, n_samples):
    """Compute p-values from t-values
    Given an array of test statistics from multiple t-tests and the (fixed)
    number of samples used for each test, compute p-values for all tests
    """
    # get p-values from test statistics using the cumulative distribution
    # function of the student distribution
    return 1 - ss.t.cdf(np.abs(tvals), df=n_samples - 1)


def pval_to_tval(pval, n_samples):
    """Conversion of a single p-value to corresponding t
    Given a desired p-value and sample size, return the corresponding t-value.
    The absolute value for the 2 sided hypothesis is returned.
    """
    return abs(ss.t.ppf(pval / 2, df=n_samples - 1))



#!/usr/bin/env python3

"""
Functions used for change detection.
cmdoret, 20200403
"""
from typing import Iterable
import scipy.sparse as sp
import numpy as np
import chromosight.utils.detection as cud
import chromosight.utils.preprocessing as cup


def median_bg(mats: Iterable[sp.spmatrix]) -> sp.spmatrix:
    """
    Given multiple sparse matrices with the same shape, format
    and nonzero coordinates, generate a background matrix made
    up of the median signal

    Parameters
    ----------
    mats : Iterable of scipy.sparse.spmatrix:
        A list of data matrices from different samples.

    Returns
    -------
    scipy.sparse.spmatrix :
        The median background matrix, where each pixel is the median of
        the corresponding pixel from all samples.
    """
    # Check that all matrices are sparse, have the same format and sparsity
    if not np.all([sp.issparse(m) for m in mats]):
        raise ValueError("Input must be a scipy sparse matrix.")
    if not np.all([m.format == mats[0].format for m in mats]):
        raise ValueError("All input matrices must be in the same format.")
    if not np.all([m.getnnz() == mats[0].getnnz() for m in mats]):
        raise ValueError("All input matrices must have the same sparsity.")
    bg = mats[0].copy()
    bg.data = np.median([m.data for m in mats], axis=0)
    return bg


def reps_bg_diff(mats: Iterable[sp.spmatrix]) -> np.ndarray:
    """
    Given multiple sample matrices, return a 1D array of all differences
    between each sample matrix's pixels and their corresponding median
    background value. All input matrices must have the same shape, nonzero
    coordinates and format.

    Parameters
    ----------
    mats : Iterable of scipy.sparse.spmatrix
        The list of data matrices from different samples.

    Returns
    -------
    numpy.ndarray of floats :
        The distribution of pixel differences to the median background. If
        there are S matrices of P nonzero pixels, this 1D array will contain
        P*S elements.
    """
    med_bg = median_bg(mats)
    diffs = [m.data - med_bg.data for m in mats]
    diffs = np.hstack(diffs)
    return diffs


def get_sse_mat(mats: Iterable[sp.spmatrix]) -> sp.spmatrix:
    """
    Given multiple matrices, return the matrix of position-wise sum of
    squared errors to median. All input matrices must have the same shape,
    nonzero coordinates and format. Output matrix will have the same format
    as inputs.

    Parameters
    ----------
    mats : Iterable of scipy.sparse.spmatrix
        The list of data matrices from different samples.

    Returns
    -------
    scipy.sparse.spmatrix :
        A sparse matrix where each nonzero pixel is the sum of squared
        difference of all samples to their median background, at the
        corresponding nonzero pixel.

    """
    bg = median_bg(mats)
    sse = bg.copy()
    se = np.array([(m.data - bg.data) ** 2 for m in mats])
    sse.data = np.sum(se, axis=0)
    return sse


def get_win_density(
    mat: sp.csr_matrix, win_size: int = 3, sym_upper: bool = False
) -> sp.csr_matrix:
    """Compute local pixel density in sparse matrices using convolution.
    The convolution is performed in 'full' mode: Computations are performed
    all the way to the edges by trimming the kernel.

    Parameters
    ----------
    mat : scipy.sparse.csr_matrix
        The input sparse matrix to convolve.
    win_size : int
        The size of the area in which to compute the proportion of nonzero
        pixels. This will be the kernel size used in convolution.
    sym_upper : bool
        Whether the input matrix is symmetric upper, in which case only the
        upper triangle is returned.


    Returns
    -------
    scipy.sparse.csr_matrix:
        The result of the convolution of the uniform kernel
        (win_size x win_size) on the binarized input matrix.
        Each value represents the proportion of nonzero pixels
        in the neighbourhood.
    """
    ms, ns = mat.shape
    # Generate a binary matrix (pixels are either empty or full)
    bin_mat = mat.copy()
    if sym_upper:
        bin_mat = sp.triu(bin_mat)
    bin_mat.data = bin_mat.data.astype(bool)
    # Adding a frame of zeros around the signal
    tmp = sp.csr_matrix((win_size - 1, ns), dtype=bool)
    bin_mat = sp.vstack([tmp, bin_mat, tmp], format=mat.format)
    tmp = sp.csr_matrix((ms + 2 * (win_size - 1), win_size - 1), dtype=bool)
    bin_mat = sp.hstack([tmp, bin_mat, tmp], format=mat.format)
    # Convolve the uniform kernel with this matrix to get the proportion of nonzero
    # pixels in each neighbourhood
    kernel = np.ones((win_size, win_size))
    win_area = win_size ** 2
    density = cud.xcorr2(bin_mat, kernel / win_area)

    # Compute convolution of uniform kernel with a frame of ones to get number
    # of missing pixels in each window.
    frame = cup.frame_missing_mask(
        sp.csr_matrix(mat.shape, dtype=bool), kernel.shape, sym_upper=sym_upper
    )
    frame = cud.xcorr2(frame, kernel).tocoo()
    # From now on, frame.data contains the number of 'present' samples.
    # (where there is at least one missing pixel)
    frame.data = win_area - frame.data

    # Adjust the proportion for values close to the border (lower denominator)
    density[frame.row, frame.col] = (
        density[frame.row, frame.col].A1 * win_area / frame.data
    )
    # Trim the frame out from the signal
    density = density[
        win_size - 1 : -win_size + 1, win_size - 1 : -win_size + 1
    ]
    if sym_upper:
        density = sp.triu(density)
    return density



import io
import os
import re

from setuptools import find_packages
from setuptools import setup


def read(filename):
    filename = os.path.join(os.path.dirname(__file__), filename)
    text_type = type(u"")
    with io.open(filename, mode="r", encoding='utf-8') as fd:
        return re.sub(text_type(r':[a-z]+:`~?(.*?)`'), text_type(r'``\1``'), fd.read())


setup(
    name="pareidolia",
    version="0.6.1",
    url="https://github.com/cmdoret/pareidolia",
    license='MIT',
    author="Cyril Matthey-Doret",
    author_email="cyril.matthey-doret@pasteur.fr",
    description="Multi-sample change detection in Hi-C patterns",
    long_description=read("README.rst"),
    packages=find_packages(exclude=('tests',)),
    install_requires=read('requirements.txt').splitlines(),
    entry_points={"console_scripts": ["pareidolia=pareidolia.cli:pareidolia_cmd"]},
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        "Intended Audience :: Science/Research",
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Operating System :: OS Independent",
    ],
)



#!/usr/bin/env python3

"""
Functions used to clean and prepare sparse matrices for detection.
cmdoret, 20200403
"""
from typing import Iterable, Tuple, Iterator, Set
import numpy as np
import scipy.sparse as sp
import chromosight.utils.preprocessing as cup


def get_common_valid_bins(
    mats: Iterable["sp.csr_matrix[float]"], n_mads: float = 5,
) -> "np.ndarray[int]":
    """
    Generates an array of valid bins indices, using the intersection
    of valid bins from all input sparse matrices. All input matrices must
    be square and have the same shape. Valid bins are defined based on their
    proportion of nonzero pixels.

    Parameters
    ----------
    mats : Iterable of sp.csr_matrix
        A list sparse matrices representing Hi-C contacts, each matrix
        represents a sample.
    n_mads : float
        A bin is considered missing if its proportion of nonzero pixels is lower
        than n_mads median absolute deviations below the median of the bin
        distribution for the whole matrix.

    Returns
    -------
    np.ndarray of ints :
        A 1D array containing the indices of valid (non-missing) bins.
    """
    common_valid = None
    for mat in mats:
        if mat.shape[0] != mat.shape[1]:
            raise NotImplementedError("Only square matrices are valid input.")
        # Get the list of valid bins in the current matrix
        valid = cup.get_detectable_bins(mat, n_mads=n_mads)
        # Initialize set of common bins with the first matrix
        if common_valid is None:
            common_valid = set(valid[0])
        # Remove elements absent from current matrix from the common set
        else:
            common_valid = common_valid.intersection(set(valid[0]))
    return np.array(list(common_valid))


def get_nnz_union(mats: Iterable["sp.csr_matrix[float]"]) -> "np.ndarray[int]":
    """
    Given a list of sparse matrices, return the union of their nonzero
    coordinates, in the form of a 2D numpy array with 1 coordinate per
    row, with 2 columns representing coordinates rows and columns.

    Parameters
    ----------
    mats : Iterable of sp.csr_matrix
        List containing the sparse matrices from each sample.

    Returns
    -------
    np.ndarray of int :
        A 2D numpy array containing the union of nonzero coordinates in the
        input sparse matrices. The array has shape Nx2 where N is the number of
        coordinates. The columns represent row and column coordinates.
    """
    try:
        # Check for input type
        if np.all([m.format == "csr" for m in mats]):
            for i, mat in enumerate(mats):
                # Use first matrix to initialize set
                if i == 0:
                    union_mat = mat.copy().astype(bool, copy=True)
                # Iteratively sum matrices
                else:
                    union_mat += mat.astype(bool, copy=True)
                union_mat.eliminate_zeros()
            # Retrieve positions of nonzero entries into an array
            all_nnz = np.ascontiguousarray(np.vstack(union_mat.nonzero()).T)
        else:
            raise ValueError("input sparse matrices must be in csr format")
    except AttributeError:
        raise TypeError("Input type must be scipy.sparse.csr_matrix")

    return all_nnz


def fill_nnz(
    mat: "sp.csr_matrix", all_nnz: "np.ndarray[int]", fill_value: float = 1e-9
) -> sp.csr_matrix:
    """
    Given an input sparse matrix and a superset of nonzero coordinates, fill the
    matrix to ensure all values in the set are stored explicitely with the
    value of fill_value.

    Parameters
    ----------
    mat : sp.csr_matrix
        The sparse matrix of a single sample.
    all_nnz : np.ndarray of ints
        A 2D array of shape Nx2, where N is the number of nonzero elements to
        fill. The columns represent the row and column coordinates of those
        elements.
    fill_value : float
        The value to use when filling the nonzero elements in the matrix. Has to
        be the same datatype as the input matrix.

    Returns
    -------
    sp.csr_matrix :
        The filled sparse matrix, where all coordinates in all_nnz have been filled
        with fill_value.
    """
    # Get the set of nonzero coordinate in the input matrix
    mat_nnz = np.ascontiguousarray(np.vstack(mat.nonzero()).T)
    out = mat.copy()
    ncols = all_nnz.shape[1]
    # Tricking numpy into treating rows as single values using a custom dtype
    # based on: https://stackoverflow.com/a/8317403/8440675
    dtype = {
        "names": [f"f{i}" for i in range(ncols)],
        "formats": ncols * [all_nnz.dtype],
    }
    # get all all_nnz coordinates that are zero in the matrix
    add_mask = np.in1d(all_nnz.view(dtype), mat_nnz.view(dtype), invert=True)
    # Replace implicit zeros by fill_value at these coordinates
    out[all_nnz[add_mask, 0], all_nnz[add_mask, 1]] = fill_value
    out.eliminate_zeros()
    return out
